@@ -2,7 +2,7 @@
 
 #==============================================================================#
 # exerb.rb
-# $Id: exerb.rb,v 1.9 2002/09/20 09:12:12 yuya Exp $
+# $Id: exerb.rb,v 1.12 2002/10/23 02:02:51 yuya Exp $
 #==============================================================================#
 
 require 'getoptlong'
@@ -41,7 +41,9 @@ module ExerbCommand
       }
     end
 
-    arcbin  = Exerb::Binary::Archive.new(arc.pack)
+    compress = !!options['compress']
+
+    arcbin  = Exerb::Binary::Archive.new(arc.pack, compress)
     corebin = Exerb::Binary::Core.new_from_file(corefile)
     exebin  = Exerb::Binary::Executable.new(arcbin, corebin)
     exebin.selfcheck
@@ -57,7 +59,7 @@ module ExerbCommand
     if Exerb::CORE.has_key?(name)
       return Exerb::CORE[name]
     else
-      raise "core name not found in exerb/config.rb : #{options['corename']}"
+      raise "core name not found in exerb/config.rb : #{name}"
     end
   end
 
@@ -77,6 +79,7 @@ module ExerbCommand
       ['--corename', '-c', GetoptLong::REQUIRED_ARGUMENT],
       ['--corepath', '-C', GetoptLong::REQUIRED_ARGUMENT],
       ['--output',   '-o', GetoptLong::REQUIRED_ARGUMENT],
+      ['--compress', '-z', GetoptLong::NO_ARGUMENT],
       ['--debug',    '-d', GetoptLong::NO_ARGUMENT],
       ['--help',     '-h', GetoptLong::NO_ARGUMENT]
     )
@@ -94,6 +97,7 @@ module ExerbCommand
     puts "  -c  --corename  specifies exerb-core name."
     puts "  -C  --corepath  specifies exerb-core file."
     puts "  -o  --output    specifies output file."
+    puts "  -z  --compress  compress the archive. (using Ruby/zlib)"
     puts "  -h  --help      display this information."
   end
 
@@ -118,9 +122,11 @@ module ExerbCommand
 
       @list.each { |type, name, path|
         case type
-        when 'kcode'   then arc.set_kcode(name)
-        when 'script'  then arc.add_script(name, File.expand_path(path, @basedir))
-        when 'library' then arc.add_dll(name, File.expand_path(path, @basedir))
+        when 'kcode'    then arc.set_kcode(name)
+        when 'corename' then raise NotImplementedError
+        when 'corepath' then raise NotImplementedError
+        when 'script'   then arc.add_script(name, File.expand_path(path, @basedir))
+        when 'library'  then arc.add_dll(name, File.expand_path(path, @basedir))
         else raise "syntax error : #{type}"
         end
       }
