@@ -1,4 +1,4 @@
-// $Id: exerb.cpp,v 1.43 2002/09/23 15:30:25 yuya Exp $
+// $Id: exerb.cpp,v 1.55 2002/10/28 15:25:52 yuya Exp $
 
 ////////////////////////////////////////////////////////////////////////////////
 
@@ -12,6 +12,10 @@
 #include "exerb.h"
 #include "resource.h"
 
+#ifdef USE_ZLIB
+#include "zlib.h"
+#endif
+
 ////////////////////////////////////////////////////////////////////////////////
 
 using namespace std;
@@ -26,6 +30,16 @@ using namespace std;
 #define GUI
 #endif
 
+#ifdef _DEBUG
+#define DEBUGMSG0(msg)          ::fprintf(stderr, msg)
+#define DEBUGMSG1(msg, p1)      ::fprintf(stderr, msg, p1)
+#define DEBUGMSG2(msg, p1, p2)  ::fprintf(stderr, msg, p1, p2)
+#else
+#define DEBUGMSG0(msg)
+#define DEBUGMSG1(msg, p1)
+#define DEBUGMSG2(msg, p1, p2)
+#endif
+
 ////////////////////////////////////////////////////////////////////////////////
 
 #ifdef CUI
@@ -36,51 +50,56 @@ int main(int argc, char** argv);
 int WINAPI WinMain(HINSTANCE hCurrentInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int iShowCmd);
 #endif
 
-static int   ExerbMain(int argc, char** argv);
-static void  ExerbInitRuby(int argc, char** argv);
-static void  ExerbInitModule();
-static void  ExerbReplaceRequire(void);
-static bool  ExerbMapping(void);
-static void  ExerbUnMapping(void);
-static PEXERB_DOS_HEADER     ExerbGetDosHeader(void);
-static PIMAGE_NT_HEADERS   ExerbGetNtHeader(void);
-static PEXERB_ARCHIVE_HEADER ExerbGetArchiveHeader(void);
-static bool  ExerbSetup(void);
-static void  ExerbSetupTable(PEXERB_ARCHIVE_HEADER pArchive);
-static void  ExerbSetupKcode(PEXERB_ARCHIVE_HEADER pArchive);
-static int   ExerbExecute(void);
-static VALUE ExerbExecuteFirstScript(void);
-static void  ExerbCleanup(void);
-static void  ExerbFinalRuby(void);
-static int   ExerbFail(void);
-static void  ExerbPrintStackTrace();
-static LRESULT CALLBACK ExerbFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam);
-static void  ExerbErrorMessage(int iResourceID);
-static void  ExerbErrorMessage(LPSTR pszMessage);
-static void  ExerbRaiseLoadError(int iResourceID);
-static VALUE ExerbRequireWithExtension(VALUE vFileName, char *pszExtension);
-static VALUE ExerbRequireWithoutExtension(VALUE vFileName);
-static void  ExerbCreateTemporaryFile(char *pszFilePathBuffer);
-static bool  ExerbGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta);
-static PIMAGE_SECTION_HEADER ExerbGetFirstSectionHeader(PIMAGE_NT_HEADERS pNtHeader);
-static PIMAGE_SECTION_HEADER ExerbGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA);
-static PIMAGE_IMPORT_DESCRIPTOR ExerbGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta);
-static void  ExerbReplaceImportDllName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSource, char* pszDestination);
-static void  ExerbReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination);
-static void  ExerbReplaceImportTable(DWORD dwOffset);
-static void  ExerbDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath);
-static HMODULE ExerbLoadDll(char *pszFilePath, VALUE vFileName);
-static VALUE ExerbGetInitFunctionName(VALUE vFileName);
-static VALUE ExerbRequireDll(VALUE vFileName, DWORD dwID);
-static VALUE ExerbRequireScript(VALUE vFileName, DWORD dwID);
-static void  ExerbGetSelfFilePath(LPSTR pszBuffer, UINT uiBufferSize);
-static void  ExerbGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize);
-static VALUE ExerbEvalString(VALUE vString, VALUE vFileName);
+static int   ExMain(int argc, char** argv);
+static void  ExInitRuby(int argc, char** argv);
+static void  ExInitModule();
+static void  ExReplaceRequire(void);
+static bool  ExMapping(void);
+static void  ExUnMapping(void);
+static PIMAGE_NT_HEADERS ExGetNtHeader(PIMAGE_DOS_HEADER pDosHeader);
+static bool  ExSetup(void);
+static void  ExSetupTable(PEXERB_ARCHIVE_HEADER pArchive);
+static void  ExSetupKcode(PEXERB_ARCHIVE_HEADER pArchive);
+static int   ExExecute(void);
+#ifdef RUBY17
+static VALUE ExExecuteFirstScript(VALUE data);
+#else
+static VALUE ExExecuteFirstScript(void);
+#endif
+static void  ExCleanup(void);
+static void  ExFinalRuby(void);
+static int   ExFail(void);
+static void  ExPrintStackTrace();
+static LRESULT CALLBACK ExFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam);
+static void  ExErrorMessage(int iResourceID);
+static void  ExErrorMessage(LPSTR pszMessage);
+static void  ExRaiseLoadError(int iResourceID);
+static VALUE ExRequireWithExtension(VALUE vFileName, char *pszExtension);
+static VALUE ExRequireWithoutExtension(VALUE vFileName);
+static void  ExCreateTemporaryFile(char *pszFilePathBuffer);
+static bool  ExGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta);
+static PIMAGE_SECTION_HEADER ExGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA);
+static PIMAGE_SECTION_HEADER ExFindSection(PIMAGE_NT_HEADERS pNtHeader, char* pszSectionName);
+static PIMAGE_IMPORT_DESCRIPTOR ExGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta);
+static void  ExReplaceImportDllName(DWORD dwDosHeader, DWORD dwImportTableDelta, PDWORD pdwNamePoolAddress, PDWORD pdwNamePoolSize, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSrc, char* pszDest);
+static void  ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination);
+static bool  ExGetSectionUnusedArea(PIMAGE_NT_HEADERS pNtHeader, char* pszName, PDWORD pdwAddress, PDWORD pdwSize);
+//static void  ExGetNamePool(PIMAGE_NT_HEADERS pNtHeader, PDWORD pdwAddress, PDWORD pdwSize);
+static void  ExReplaceImportTable(DWORD dwOffset);
+static void  ExDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath);
+static HMODULE ExLoadDll(char *pszFilePath, VALUE vFileName);
+static VALUE ExGetInitFunctionName(VALUE vFileName);
+static VALUE ExRequireDll(VALUE vFileName, DWORD dwID);
+static VALUE ExRequireScript(VALUE vFileName, DWORD dwID);
+static void  ExGetSelfFilePath(LPSTR pszBuffer, UINT uiBufferSize);
+static void  ExGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize);
+static VALUE ExEvalString(VALUE vString, VALUE vFileName);
+static VALUE ExRequire(VALUE valueObject, VALUE valueFileName);
 
-static VALUE ExerbRequire(VALUE valueObject, VALUE valueFileName);
 static VALUE rb_exerb_s_runtime_p(VALUE self);
 
-VALUE WINAPI rbxrequire(const char *fname);
+VALUE WINAPI ex_require(const char *fname);
+VALUE WINAPI ex_f_require(VALUE obj, VALUE fname);
 
 ////////////////////////////////////////////////////////////////////////////////
 
@@ -112,13 +131,15 @@ static map<DWORD, DWORD> g_ScriptSizeTable;
 static map<DWORD, DWORD> g_DllOffsetTable;
 static map<DWORD, DWORD> g_DllSizeTable;
 
+static char g_szPhiSoFileName[MAX_PATH] = "";
+
 ////////////////////////////////////////////////////////////////////////////////
 
 #ifdef CUI
 int
 main(int argc, char** argv)
 {
-	return ExerbMain(argc, argv);
+	return ExMain(argc, argv);
 }
 #endif
 
@@ -126,35 +147,35 @@ main(int argc, char** argv)
 int WINAPI
 WinMain(HINSTANCE hCurrentInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int iShowCmd)
 {
-	return ExerbMain(0, NULL);
+	return ExMain(0, NULL);
 }
 #endif
 
 ////////////////////////////////////////////////////////////////////////////////
 
 static int
-ExerbMain(int argc, char** argv)
+ExMain(int argc, char** argv)
 {
 	int iResultCode = 1;
 
-	::ExerbInitRuby(argc, argv);
-	::ExerbInitModule();
-	::ExerbReplaceRequire();
+	::ExInitRuby(argc, argv);
+	::ExInitModule();
+	::ExReplaceRequire();
 	
-	if ( ::ExerbMapping() ) {
-		if ( ::ExerbSetup() ) {
-			iResultCode = ::ExerbExecute();
-			::ExerbFinalRuby();
-			::ExerbCleanup();
+	if ( ::ExMapping() ) {
+		if ( ::ExSetup() ) {
+			iResultCode = ::ExExecute();
+			::ExFinalRuby();
+			::ExCleanup();
 		}
-		::ExerbUnMapping();
+		::ExUnMapping();
 	}
 
 	return iResultCode;
 }
 
 static void
-ExerbInitRuby(int argc, char** argv)
+ExInitRuby(int argc, char** argv)
 {
 	::NtInitialize(&argc, &argv);
 	::ruby_init();
@@ -162,91 +183,93 @@ ExerbInitRuby(int argc, char** argv)
 }
 
 static void
-ExerbInitModule()
+ExInitModule()
 {
 	VALUE mExerb = ::rb_define_module("Exerb");
 	::rb_define_singleton_method(mExerb, "runtime?", (RUBYFUNC)rb_exerb_s_runtime_p, 0);
 }
 
 static void
-ExerbReplaceRequire(void)
+ExReplaceRequire(void)
 {
 	::rb_alias(rb_mKernel, ::rb_intern("__old_require__"), ::rb_intern("require"));
-	::rb_define_global_function("require", (RUBYFUNC)ExerbRequire, 1);
+	::rb_define_global_function("require", (RUBYFUNC)ExRequire, 1);
 }
 
 static bool
-ExerbMapping(void)
+ExMapping(void)
 {
-	char szFilePath[MAX_PATH] = "";
-	::ExerbGetSelfFilePath(szFilePath, sizeof(szFilePath));
-
-	PEXERB_DOS_HEADER pDosHeader = ::ExerbGetDosHeader();
+	PEXERB_DOS_HEADER pDosHeader = (PEXERB_DOS_HEADER)::GetModuleHandle(NULL);
 	if ( !pDosHeader->OffsetToArchive || !pDosHeader->SizeOfArchive ) {
-	    ::ExerbErrorMessage(IDS_HEADER_NOT_FOUND);
+	    ::ExErrorMessage(IDS_HEADER_NOT_FOUND);
 	    return false;
 	}
 
-	g_pArchive = new char[pDosHeader->SizeOfArchive];
+	int nArchiveSize = (pDosHeader->SizeOfCompressedArchive > 0 ? pDosHeader->SizeOfCompressedArchive : pDosHeader->SizeOfArchive);
+	g_pArchive = new char[nArchiveSize];
+
+	char szFilePath[MAX_PATH] = "";
+	::ExGetSelfFilePath(szFilePath, sizeof(szFilePath));
 
 	HANDLE hFile = ::CreateFile(szFilePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
 	if ( hFile == INVALID_HANDLE_VALUE ) {
-		::ExerbErrorMessage(IDS_FAIL_CREATEFILE);
+		::ExErrorMessage(IDS_FAIL_CREATEFILE);
 		return false;
 	}
 
 	DWORD dwRead = 0;
 	::SetFilePointer(hFile, pDosHeader->OffsetToArchive, 0, FILE_BEGIN);
-	::ReadFile(hFile, g_pArchive, pDosHeader->SizeOfArchive, &dwRead, NULL);
+	::ReadFile(hFile, g_pArchive, nArchiveSize, &dwRead, NULL);
 
 	::CloseHandle(hFile);
 
+	if ( pDosHeader->SizeOfCompressedArchive > 0 ) {
+#ifdef USE_ZLIB
+		DWORD dwOriginalSize   = pDosHeader->SizeOfArchive;
+		DWORD dwCompressedSize = pDosHeader->SizeOfCompressedArchive;
+		void* pUncompressedArchive = new char[dwOriginalSize];
+		::uncompress((BYTE*)pUncompressedArchive, &dwOriginalSize, (BYTE*)g_pArchive, dwCompressedSize);
+		delete[] g_pArchive;
+		g_pArchive = pUncompressedArchive;
+#else
+	    ::ExErrorMessage("The compressed archive isn't supported by the core.");
+	    return false;
+#endif
+	}
+
 	return true;
 }
 
 static void
-ExerbUnMapping(void)
+ExUnMapping(void)
 {
 	delete[] g_pArchive;
 }
 
-static PEXERB_DOS_HEADER
-ExerbGetDosHeader(void)
-{
-	return (PEXERB_DOS_HEADER)::GetModuleHandle(NULL);
-}
-
 static PIMAGE_NT_HEADERS
-ExerbGetNtHeader(void)
+ExGetNtHeader(PIMAGE_DOS_HEADER pDosHeader)
 {
-	PEXERB_DOS_HEADER pDosHeader = ::ExerbGetDosHeader();
-	return (PIMAGE_NT_HEADERS)((DWORD)pDosHeader + pDosHeader->OffsetToNewHeader);
-}
-
-static PEXERB_ARCHIVE_HEADER
-ExerbGetArchiveHeader(void)
-{
-	return (PEXERB_ARCHIVE_HEADER)g_pArchive;
+	return (PIMAGE_NT_HEADERS)((DWORD)pDosHeader + pDosHeader->e_lfanew);
 }
 
 static bool
-ExerbSetup(void)
+ExSetup(void)
 {
-	PEXERB_ARCHIVE_HEADER pArchive = ::ExerbGetArchiveHeader();
+	PEXERB_ARCHIVE_HEADER pArchive = (PEXERB_ARCHIVE_HEADER)g_pArchive;
 
 	if ( pArchive->NumberOfScript == 0 ) {
-		::ExerbErrorMessage(IDS_SCRIPT_NOT_FOUND);
+		::ExErrorMessage(IDS_SCRIPT_NOT_FOUND);
 		return false;
 	}
 
-	ExerbSetupTable(pArchive);
-	ExerbSetupKcode(pArchive);
+	::ExSetupTable(pArchive);
+	::ExSetupKcode(pArchive);
 
 	return true;
 }
 
 static void
-ExerbSetupTable(PEXERB_ARCHIVE_HEADER pArchive)
+ExSetupTable(PEXERB_ARCHIVE_HEADER pArchive)
 {
 	DWORD i = 0;
 
@@ -276,7 +299,7 @@ ExerbSetupTable(PEXERB_ARCHIVE_HEADER pArchive)
 }
 
 static void
-ExerbSetupKcode(PEXERB_ARCHIVE_HEADER pArchive)
+ExSetupKcode(PEXERB_ARCHIVE_HEADER pArchive)
 {
 	switch ( pArchive->Options.Kcode ) {
 	case EXERB_OPTIONS_KCODE_NONE: ::rb_set_kcode("n"); break; // NONE
@@ -287,33 +310,37 @@ ExerbSetupKcode(PEXERB_ARCHIVE_HEADER pArchive)
 }
 
 static int
-ExerbExecute(void)
+ExExecute(void)
 {
 	int state = 0;
 
-	::rb_protect(ExerbExecuteFirstScript, 0, &state);
+	::rb_protect(ExExecuteFirstScript, 0, &state);
 
 	if ( state ) {
-		return ::ExerbFail();
+		return ::ExFail();
 	} else {
 		return 0;
 	}
 }
 
 static VALUE
-ExerbExecuteFirstScript(void)
+#ifdef RUBY17
+ExExecuteFirstScript(VALUE data)
+#else
+ExExecuteFirstScript(void)
+#endif
 {
 	PEXERB_SCRIPT_HEADER pScriptHeader = (PEXERB_SCRIPT_HEADER)g_dwBaseOfScriptTable;
 	VALUE vScirptSource = ::rb_str_new((char*)(g_dwBaseOfScriptTable + pScriptHeader->Offset), pScriptHeader->Size);
 	VALUE vScritpName   = ::rb_str_new2(g_Id2NameTable[pScriptHeader->NameID].c_str());
 	::ruby_script(STR2CSTR(vScritpName));
-	::ExerbEvalString(vScirptSource, vScritpName);
+	::ExEvalString(vScirptSource, vScritpName);
 
 	return Qnil;
 }
 
 static void
-ExerbCleanup(void)
+ExCleanup(void)
 {
 	g_Name2IdTable.clear();
 	g_Id2NameTable.clear();
@@ -322,7 +349,7 @@ ExerbCleanup(void)
 	g_DllOffsetTable.clear();
 	g_DllSizeTable.clear();
 
-	for ( int i = 0; i < g_DllTable.size(); i++ ) {
+	for ( int i = g_DllTable.size() - 1; i >= 0 ; i-- ) {
 		DLL_ENTRY* entry = g_DllTable[i];
 		::FreeLibrary(entry->hModule);
 		::DeleteFile(entry->pszFilePath);
@@ -334,34 +361,34 @@ ExerbCleanup(void)
 }
 
 static void
-ExerbFinalRuby(void)
+ExFinalRuby(void)
 {
 	::ruby_finalize();
 }
 
 static int
-ExerbFail(void)
+ExFail(void)
 {
 	if ( ruby_errinfo ) {
 		if ( ::rb_obj_is_kind_of(ruby_errinfo, rb_eSystemExit) == Qtrue ) {
 			return 0;
 		} else {
 #ifdef GUI
-			::DialogBox(::GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_EXCEPTION), NULL, (DLGPROC)::ExerbFailDialogProc);
+			::DialogBox(::GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_EXCEPTION), NULL, (DLGPROC)::ExFailDialogProc);
 #endif
 #ifdef CUI
-			::ExerbPrintStackTrace();
+			::ExPrintStackTrace();
 #endif
 			return 1;
 		}
 	} else {
-		::ExerbErrorMessage(IDS_CANNOT_OBTAIN_EXCEPTION);
+		::ExErrorMessage(IDS_CANNOT_OBTAIN_EXCEPTION);
 		return 1;
 	}
 }
 
 static void
-ExerbPrintStackTrace()
+ExPrintStackTrace()
 {
 	VALUE type       = ::rb_funcall(::rb_funcall(ruby_errinfo, ::rb_intern("type"), 0), ::rb_intern("name"), 0);
 	VALUE message    = ::rb_funcall(ruby_errinfo, ::rb_intern("message"), 0);
@@ -376,7 +403,7 @@ ExerbPrintStackTrace()
 }
 
 static LRESULT CALLBACK
-ExerbFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
+ExFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
 {
 	switch ( uiMessage ) {
 		case WM_INITDIALOG:
@@ -397,7 +424,7 @@ ExerbFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
 				char szSelfFileName[MAX_PATH]      = "";
 				char szWindowTitleFormat[MAX_PATH] = "";
 				char szWindowTitle[MAX_PATH]       = "";
-				::ExerbGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
+				::ExGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
 				::GetWindowText(hWnd, szWindowTitleFormat, sizeof(szWindowTitleFormat));
 				::wsprintf(szWindowTitle, szWindowTitleFormat, szSelfFileName);
 				::SetWindowText(hWnd, szWindowTitle);
@@ -428,19 +455,19 @@ ExerbFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
 }
 
 static void
-ExerbErrorMessage(int iResourceID)
+ExErrorMessage(int iResourceID)
 {
 	char message[1024] = "";
 	::LoadString(::GetModuleHandle(NULL), iResourceID, message, sizeof(message));
-	::ExerbErrorMessage(message);
+	::ExErrorMessage(message);
 }
 
 static void
-ExerbErrorMessage(LPSTR pszMessage)
+ExErrorMessage(LPSTR pszMessage)
 {
 #ifdef GUI
 	char szSelfFileName[MAX_PATH] = "";
-	::ExerbGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
+	::ExGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
 	::MessageBox(NULL, pszMessage, szSelfFileName, MB_ICONERROR);
 #endif
 #ifdef CUI
@@ -449,7 +476,7 @@ ExerbErrorMessage(LPSTR pszMessage)
 }
 
 static void
-ExerbRaiseLoadError(int iResourceID)
+ExRaiseLoadError(int iResourceID)
 {
 	char message[1024] = "";
 	::LoadString(::GetModuleHandle(NULL), iResourceID, message, sizeof(message));
@@ -457,15 +484,15 @@ ExerbRaiseLoadError(int iResourceID)
 }
 
 static VALUE
-ExerbRequireWithExtension(VALUE vFileName, char *pszExtension)
+ExRequireWithExtension(VALUE vFileName, char *pszExtension)
 {
 	DWORD dwID = g_Name2IdTable[STR2CSTR(vFileName)];
 
 	if ( dwID ) {
 		if ( ::stricmp(".rb", pszExtension) == 0 ) {
-			return ::ExerbRequireScript(vFileName, dwID);
+			return ::ExRequireScript(vFileName, dwID);
 		} else if ( ::stricmp(".so", pszExtension) == 0 || ::stricmp(".dll", pszExtension) == 0 ) {
-			return ::ExerbRequireDll(vFileName, dwID);
+			return ::ExRequireDll(vFileName, dwID);
 		} else {
 			return Qfalse;
 		}
@@ -475,7 +502,7 @@ ExerbRequireWithExtension(VALUE vFileName, char *pszExtension)
 }
 
 static VALUE
-ExerbRequireWithoutExtension(VALUE vFileName)
+ExRequireWithoutExtension(VALUE vFileName)
 {
 	VALUE vFileNameRB  = ::rb_str_concat(::rb_str_dup(vFileName), ::rb_str_new2(".rb"));
 	VALUE vFileNameSO  = ::rb_str_concat(::rb_str_dup(vFileName), ::rb_str_new2(".so"));
@@ -493,18 +520,18 @@ ExerbRequireWithoutExtension(VALUE vFileName)
 */
 
 	if ( idRB ) {
-		return ::ExerbRequireScript(vFileNameRB, idRB);
+		return ::ExRequireScript(vFileNameRB, idRB);
 	} else if ( idSO ) {
-		return ::ExerbRequireDll(vFileNameSO, idSO);
+		return ::ExRequireDll(vFileNameSO, idSO);
 	} else if ( idDLL ) {
-		return ::ExerbRequireDll(vFileNameDLL, idDLL);
+		return ::ExRequireDll(vFileNameDLL, idDLL);
 	} else {
 		return Qfalse;
 	}
 }
 
 static void
-ExerbCreateTemporaryFile(char *pszFilePathBuffer)
+ExCreateTemporaryFile(char *pszFilePathBuffer)
 {
 	char szTemporaryDirectoryPath[MAX_PATH] = "";
 	::GetTempPath(sizeof(szTemporaryDirectoryPath), szTemporaryDirectoryPath);
@@ -512,13 +539,13 @@ ExerbCreateTemporaryFile(char *pszFilePathBuffer)
 }
 
 static bool
-ExerbGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta)
+ExGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta)
 {
 	PIMAGE_DOS_HEADER pDosHeader = (PIMAGE_DOS_HEADER)dwDosHeader;
-	PIMAGE_NT_HEADERS pNtHeader  = (PIMAGE_NT_HEADERS)((DWORD)pDosHeader + pDosHeader->e_lfanew);
+	PIMAGE_NT_HEADERS pNtHeader  = ::ExGetNtHeader(pDosHeader);
 	DWORD dwImportTableRVA = pNtHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
 
-	PIMAGE_SECTION_HEADER pSection = ::ExerbGetEnclosingSectionHeader(pNtHeader, dwImportTableRVA);
+	PIMAGE_SECTION_HEADER pSection = ::ExGetEnclosingSectionHeader(pNtHeader, dwImportTableRVA);
 	if ( pSection ) {
 		*pdwDelta   = pSection->VirtualAddress - pSection->PointerToRawData;
 		*pdwAddress = dwImportTableRVA - *pdwDelta;
@@ -531,15 +558,9 @@ ExerbGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta)
 }
 
 static PIMAGE_SECTION_HEADER
-ExerbGetFirstSectionHeader(PIMAGE_NT_HEADERS pNtHeader)
+ExGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA)
 {
-	return (PIMAGE_SECTION_HEADER)((DWORD)pNtHeader + sizeof(IMAGE_NT_HEADERS));
-}
-
-static PIMAGE_SECTION_HEADER
-ExerbGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA)
-{
-	PIMAGE_SECTION_HEADER pSection = ExerbGetFirstSectionHeader(pNtHeader);
+	PIMAGE_SECTION_HEADER pSection = IMAGE_FIRST_SECTION(pNtHeader);
 
 	for ( int i = 0; i < pNtHeader->FileHeader.NumberOfSections; i++, pSection++ ) {
 		if ( (dwRVA >= pSection->VirtualAddress) && (dwRVA < (pSection->VirtualAddress + pSection->Misc.VirtualSize)) ) {
@@ -550,11 +571,25 @@ ExerbGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA)
 	return NULL;
 }
 
+static PIMAGE_SECTION_HEADER
+ExFindSection(PIMAGE_NT_HEADERS pNtHeader, char* pszSectionName)
+{
+	PIMAGE_SECTION_HEADER pSection = IMAGE_FIRST_SECTION(pNtHeader);
+
+	for ( int i = 0; i < pNtHeader->FileHeader.NumberOfSections; i++, pSection++ ) {
+		if ( ::strnicmp(pszSectionName, (char*)pSection->Name, 8) == 0 ) {
+			return pSection;
+		}
+	}
+
+	return NULL;
+}
+
 static PIMAGE_IMPORT_DESCRIPTOR
-ExerbGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta)
+ExGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta)
 {
 	DWORD dwImportTableAddress = 0;
-	if ( ::ExerbGetImportTable(dwDosHeader, &dwImportTableAddress, pdwImportTableDelta) ) {
+	if ( ::ExGetImportTable(dwDosHeader, &dwImportTableAddress, pdwImportTableDelta) ) {
 		DWORD dwImportTableBase = dwDosHeader + dwImportTableAddress;
 		return (PIMAGE_IMPORT_DESCRIPTOR)dwImportTableBase;
 	} else {
@@ -563,23 +598,34 @@ ExerbGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta)
 }
 
 static void
-ExerbReplaceImportDllName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSource, char* pszDestination)
+ExReplaceImportDllName(DWORD dwDosHeader, DWORD dwImportTableDelta, PDWORD pdwNamePoolAddress, PDWORD pdwNamePoolSize, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSrc, char* pszDest)
 {
+	DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
+	DWORD dwDestLength   = ::strlen(pszDest);
+
 	for ( PIMAGE_IMPORT_DESCRIPTOR pDescriptor = pFirstDescriptor; pDescriptor->Name; pDescriptor++ ) {
 		char *pszName = (char*)(dwOffsetOfName + pDescriptor->Name);
 
-		if ( ::stricmp(pszName, pszSource) == 0 ) {
-			if ( ::strlen(pszDestination) <= ::strlen(pszName) ) {
-				::strcpy(pszName, pszDestination);
+		if ( ::stricmp(pszName, pszSrc) == 0 ) {
+			if ( dwDestLength <= ::strlen(pszName) ) {
+				::strcpy(pszName, pszDest);
+			} else if ( dwDestLength + 1 <= *pdwNamePoolSize ) {
+				DWORD dwAddress = *pdwNamePoolAddress - dwDestLength - 1;
+				pDescriptor->Name = dwAddress + dwImportTableDelta;
+
+				::memcpy((void*)(dwDosHeader + dwAddress), pszDest, dwDestLength);
+
+				*pdwNamePoolAddress -= dwDestLength + 1;
+				*pdwNamePoolSize    -= dwDestLength + 1;
 			} else {
-				::rb_raise(rb_eLoadError, "Fail to modify the import table.\nThe excutable file name is too long.");
+				::rb_raise(rb_eLoadError, "Fail to modify the import table. exe/dll file name is too long.");
 			}
 		}
 	}
 }
 
 static void
-ExerbReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination)
+ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination)
 {
 	for ( PIMAGE_IMPORT_DESCRIPTOR pDescriptor = pFirstDescriptor; pDescriptor->Name; pDescriptor++ ) {
 		char *pszName = (char*)(dwOffsetOfName + pDescriptor->Name);
@@ -614,30 +660,84 @@ ExerbReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pF
 	}
 }
 
+static bool
+ExGetSectionUnusedArea(PIMAGE_NT_HEADERS pNtHeader, char* pszName, PDWORD pdwAddress, PDWORD pdwSize)
+{
+	PIMAGE_SECTION_HEADER pSection = ::ExFindSection(pNtHeader, pszName);
+
+	if ( pSection ) {
+		*pdwAddress = pSection->PointerToRawData + pSection->SizeOfRawData;
+		*pdwSize    = pSection->SizeOfRawData    - pSection->Misc.VirtualSize;
+		return true;
+	} else {
+		*pdwAddress = 0;
+		*pdwSize    = 0;
+		return false;
+	}
+}
+
+/*
 static void
-ExerbReplaceImportTable(DWORD dwOffset)
+ExGetNamePool(PIMAGE_NT_HEADERS pNtHeader, PDWORD pdwAddress, PDWORD pdwSize)
+{
+	char* ppszSections[] = {".idata", ".edata", ".rsrc"};
+	int   nSections      = 3;
+	DWORD dwAddress = 0;
+	DWORD dwSize    = 0;
+
+	for ( int i = 0; i < nSections; i++ ) {
+		if ( ::ExGetSectionUnusedArea(pNtHeader, ppszSections[i], &dwAddress, &dwSize) ) {
+			if ( *pdwSize < dwSize ) {
+				*pdwAddress = dwAddress;
+				*pdwSize    = dwSize;
+			}
+		}
+	}
+}
+*/
+
+static void
+ExReplaceImportTable(DWORD dwOffset)
 {
 	DWORD dwDosHeader = dwOffset;
 
 	DWORD dwImportTableDelta = 0;
-	PIMAGE_IMPORT_DESCRIPTOR pDescriptor = ExerbGetFirstImportDescriptor(dwDosHeader, &dwImportTableDelta);
+	PIMAGE_IMPORT_DESCRIPTOR pDescriptor = ::ExGetFirstImportDescriptor(dwDosHeader, &dwImportTableDelta);
 	if ( !pDescriptor ) {
-		::ExerbRaiseLoadError(IDS_FAIL_MODIFY_IMPORT_TABLE);
+		::ExRaiseLoadError(IDS_FAIL_MODIFY_IMPORT_TABLE);
 	}
 
+	DWORD dwNamePoolAddress = 0;
+	DWORD dwNamePoolSize    = 0;
+	PIMAGE_NT_HEADERS     pNtHeader      = ::ExGetNtHeader((PIMAGE_DOS_HEADER)dwDosHeader);
+
+	::ExGetSectionUnusedArea(pNtHeader, ".idata", &dwNamePoolAddress, &dwNamePoolSize);
+	//::ExGetNamePool(pNtHeader, &dwNamePoolAddress, &dwNamePoolSize);
+
 	char szSelfFileName[MAX_PATH] = "";
-	::ExerbGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
+	::ExGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
+
+#ifdef RUBY17
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "msvcrt-ruby17.dll",  szSelfFileName);
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "cygwin-ruby17.dll",  szSelfFileName);
+#else
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "mswin32-ruby16.dll", szSelfFileName);
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "mingw32-ruby16.dll", szSelfFileName);
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "cygwin-ruby16.dll",  szSelfFileName);
+#endif
+	::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "ruby.exe",           szSelfFileName);
+
+	if ( ::strlen(g_szPhiSoFileName) > 0 ) {
+		::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "phi.so", g_szPhiSoFileName);
+	}
 
 	DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
-	ExerbReplaceImportDllName(dwOffsetOfName, pDescriptor, "mswin32-ruby16.dll", szSelfFileName);
-	ExerbReplaceImportDllName(dwOffsetOfName, pDescriptor, "mingw32-ruby16.dll", szSelfFileName);
-	ExerbReplaceImportDllName(dwOffsetOfName, pDescriptor, "cygwin-ruby16.dll",  szSelfFileName);
-
-	ExerbReplaceImportFunctionName(dwOffsetOfName, pDescriptor, szSelfFileName, "rb_require", "rbxrequire");
+	::ExReplaceImportFunctionName(dwOffsetOfName, pDescriptor, szSelfFileName, "rb_require",   "ex_require");
+	::ExReplaceImportFunctionName(dwOffsetOfName, pDescriptor, szSelfFileName, "rb_f_require", "ex_f_require");
 }
 
 static void
-ExerbDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath)
+ExDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath)
 {
 	void  *pvBuffer = (void*)dwOffset;
 	DWORD dwWritten = 0;
@@ -648,35 +748,43 @@ ExerbDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath)
 }
 
 static HMODULE
-ExerbLoadDll(char *pszFilePath, VALUE vFileName)
+ExLoadDll(char *pszFilePath, VALUE vFileName)
 {
+	DEBUGMSG2("ExLoadDll('%s', '%s')\n", pszFilePath, STR2CSTR(vFileName));
+
 	HMODULE hModule = ::LoadLibraryEx(pszFilePath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
 	DWORD   dwError = ::GetLastError();
 
 	if ( hModule ) {
-		VALUE vInitFunctionName = ::ExerbGetInitFunctionName(vFileName);
+		VALUE vInitFunctionName = ::ExGetInitFunctionName(vFileName);
 		DWORD dwInitProc = (DWORD)::GetProcAddress(hModule, STR2CSTR(vInitFunctionName));
 
 		if ( dwInitProc ) {
+			DEBUGMSG0("  module info\n");
+			DEBUGMSG1("    handle = 0x%08X\n", hModule);
+			DEBUGMSG0("  initialize function info\n");
+			DEBUGMSG1("    name    = '%s'\n", STR2CSTR(vInitFunctionName));
+			DEBUGMSG1("    address = 0x%08X\n", dwInitProc);
+			DEBUGMSG0("  call initialize function...\n");
 			INITPROC fpInitProc = (INITPROC)dwInitProc;
 			(*fpInitProc)();
+			DEBUGMSG0("  succeeded\n");
 			return hModule;
 		} else {
 			::FreeLibrary(hModule);
 			::DeleteFile(pszFilePath);
-			::ExerbRaiseLoadError(IDS_FAIL_GETPROCADDRESS);
+			::ExRaiseLoadError(IDS_FAIL_GETPROCADDRESS);
 			return NULL;
 		}
 	} else {
 		::DeleteFile(pszFilePath);
-		::rb_raise(rb_eLoadError, "Fail to read an extension library.\nLoadLibraryEx failed. \n The error No. was [%i].", dwError);
-		//::rb_raise(rb_eLoadError, "gCu̓ǂݍ݂Ɏs܂B\nLoadLibraryEx֐̎sɎs܂B\nG[ԍ[%i]łB", dwError);
+		::rb_raise(rb_eLoadError, "LoadLibraryEx() failed. [Error:%i].", dwError);
 		return NULL;
 	}
 }
 
 static VALUE
-ExerbGetInitFunctionName(VALUE vFileName)
+ExGetInitFunctionName(VALUE vFileName)
 {
 	static VALUE id_basename = ::rb_intern("basename");
 	static VALUE id_sub      = ::rb_intern("sub");
@@ -691,20 +799,20 @@ ExerbGetInitFunctionName(VALUE vFileName)
 }
 
 static VALUE
-ExerbRequireDll(VALUE vFileName, DWORD dwID)
+ExRequireDll(VALUE vFileName, DWORD dwID)
 {
 	DWORD dwOffset = g_DllOffsetTable[dwID];
 	DWORD dwSize   = g_DllSizeTable[dwID];
 
 	if ( dwOffset && dwSize ) {
-		::ExerbReplaceImportTable(dwOffset);
+		::ExReplaceImportTable(dwOffset);
 
 		char szTemporaryFilePath[MAX_PATH] = "";
-		::ExerbCreateTemporaryFile(szTemporaryFilePath);
+		::ExCreateTemporaryFile(szTemporaryFilePath);
 
-		::ExerbDllWriteToFile(dwOffset, dwSize, szTemporaryFilePath);
+		::ExDllWriteToFile(dwOffset, dwSize, szTemporaryFilePath);
 
-		HMODULE hModule = ::ExerbLoadDll(szTemporaryFilePath, vFileName);
+		HMODULE hModule = ::ExLoadDll(szTemporaryFilePath, vFileName);
 		if ( hModule ) {
 			DLL_ENTRY* entry = new DLL_ENTRY;
 			entry->hModule     = hModule;
@@ -712,6 +820,14 @@ ExerbRequireDll(VALUE vFileName, DWORD dwID)
 			::strcpy(entry->pszFilePath, szTemporaryFilePath);
 			g_DllTable.push_back(entry);
 
+			VALUE basename = ::rb_funcall(rb_cFile, ::rb_intern("basename"), 1, vFileName);
+			if ( ::stricmp(STR2CSTR(basename), "phi.so") == 0 ) {
+				char szFullPath[MAX_PATH] = "";
+				char* pszFileName = NULL;
+				::GetFullPathName(szTemporaryFilePath, sizeof(szFullPath), szFullPath, &pszFileName);
+				::strcpy(g_szPhiSoFileName, pszFileName);
+			}
+
 			return Qtrue;
 		} else {
 			return Qfalse;
@@ -722,41 +838,41 @@ ExerbRequireDll(VALUE vFileName, DWORD dwID)
 }
 
 static VALUE
-ExerbRequireScript(VALUE vFileName, DWORD dwID)
+ExRequireScript(VALUE vFileName, DWORD dwID)
 {
 	DWORD dwOffset = g_ScriptOffsetTable[dwID];
 	DWORD dwSize   = g_ScriptSizeTable[dwID];
 
 	if ( dwOffset && dwSize ) {
 		VALUE vScript = ::rb_str_new((char*)dwOffset, dwSize);
-		::ExerbEvalString(vScript, vFileName);
+		::ExEvalString(vScript, vFileName);
 		return Qtrue;
 	} else {
-		::ExerbRaiseLoadError(IDS_CANNOT_EXECUTE_SCRIPT);
+		::ExRaiseLoadError(IDS_CANNOT_EXECUTE_SCRIPT);
 		return Qfalse;
 	}
 }
 
 static void
-ExerbGetSelfFilePath(LPSTR pszBuffer, UINT uiBufferSize)
+ExGetSelfFilePath(LPSTR pszBuffer, UINT uiBufferSize)
 {
 	::GetModuleFileName(NULL, pszBuffer, uiBufferSize);
 }
 
 static void
-ExerbGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize)
+ExGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize)
 {
 	char szSelfFilePath[MAX_PATH]     = "";
 	char szSelfFilePathTemp[MAX_PATH] = "";
 	char *pszSelfFileName            = NULL;
 
-	::ExerbGetSelfFilePath(szSelfFilePath, sizeof(szSelfFilePath));
+	::ExGetSelfFilePath(szSelfFilePath, sizeof(szSelfFilePath));
 	::GetFullPathName(szSelfFilePath, sizeof(szSelfFilePathTemp), szSelfFilePathTemp, &pszSelfFileName);
 	::strncpy(pszBuffer, pszSelfFileName, uiBufferSize);
 }
 
 static VALUE
-ExerbEvalString(VALUE vString, VALUE vFileName)
+ExEvalString(VALUE vString, VALUE vFileName)
 {
 	static VALUE id_eval  = ::rb_intern("eval");
 	static VALUE vBinding = ::rb_const_get(rb_mKernel, ::rb_intern("TOPLEVEL_BINDING"));
@@ -765,7 +881,7 @@ ExerbEvalString(VALUE vString, VALUE vFileName)
 }
 
 static VALUE
-ExerbRequire(VALUE vObject, VALUE vFileName)
+ExRequire(VALUE vObject, VALUE vFileName)
 {
 	VALUE vFeatures = ::rb_gv_get("$\"");
 	if ( ::rb_ary_includes(vFeatures, vFileName) == Qtrue ) {
@@ -776,12 +892,12 @@ ExerbRequire(VALUE vObject, VALUE vFileName)
 	char *pszExtension = ::strrchr(pszFileName, '.');
 
 	if ( pszExtension ) {
-		if ( ::ExerbRequireWithExtension(vFileName, pszExtension) == Qtrue ) {
+		if ( ::ExRequireWithExtension(vFileName, pszExtension) == Qtrue ) {
 			::rb_ary_push(vFeatures, vFileName);
 			return Qtrue;
 		}
 	} else {
-		if ( ::ExerbRequireWithoutExtension(vFileName) == Qtrue ) {
+		if ( ::ExRequireWithoutExtension(vFileName) == Qtrue ) {
 			::rb_ary_push(vFeatures, vFileName);
 			return Qtrue;
 		}
@@ -791,16 +907,29 @@ ExerbRequire(VALUE vObject, VALUE vFileName)
 	return ::rb_funcall(rb_mKernel, id_require, 1, vFileName);
 }
 
+////////////////////////////////////////////////////////////////////////////////
+
 static VALUE
 rb_exerb_s_runtime_p(VALUE self)
 {
 	return Qtrue;
 }
 
+////////////////////////////////////////////////////////////////////////////////
+
 VALUE WINAPI
-rbxrequire(const char *fname)
+ex_require(const char *fname)
 {
-	return ::ExerbRequire(Qnil, ::rb_str_new2(fname));
+	DEBUGMSG1("ex_require('%s')\n", fname);
+	return ::ExRequire(Qnil, ::rb_str_new2(fname));
+}
+
+VALUE WINAPI
+ex_f_require(VALUE obj, VALUE fname)
+{
+	DEBUGMSG1("ex_f_require('%s')\n", STR2CSTR(fname));
+    Check_SafeStr(fname);
+	return ::ExRequire(Qnil, fname);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
